#include <linux/mm.h>
#include <linux/nmi.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>
#include <asm/mmu_context.h>
#include <linux/kernel_stat.h>
#include <linux/completion.h>
#include <linux/low-latency.h>

void set_running_and_schedule()
{
	if (current->state != TASK_RUNNING)
		__set_current_state(TASK_RUNNING);
	schedule();
}


int ll_copy_to_user(void *to_user, const void *from, unsigned long len)
{
	while (len) {
		unsigned long n_to_copy = len;
		unsigned long remainder;

		if (n_to_copy > 4096)
			n_to_copy = 4096;
		remainder = copy_to_user(to_user, from, n_to_copy);
		if (remainder)
			return remainder + len;
		to_user = ((char *)to_user) + n_to_copy;
		from = ((char *)from) + n_to_copy;
		len -= n_to_copy;
		conditional_schedule();
	}
	return 0;
}

int ll_copy_from_user(void *to, const void *from_user, unsigned long len)
{
	while (len) {
		unsigned long n_to_copy = len;
		unsigned long remainder;

		if (n_to_copy > 4096)
			n_to_copy = 4096;
		remainder = copy_from_user(to, from_user, n_to_copy);
		if (remainder)
			return remainder + len;
		to = ((char *)to) + n_to_copy;
		from_user = ((char *)from_user) + n_to_copy;
		len -= n_to_copy;
		conditional_schedule();
	}
	return 0;
}

